"""Generate RaBe Content Reference Idenitifier Spcification (CRID) Identifiers.

* [`cridlib.get(timestamp=None, fragment='')`](./get/#cridlib.get.get)
* [`cridlib.parse(value)`](./parse/#gridlib.parse.parse)
"""

from .get import get
from .lib import CRIDError
from .parse import parse

__all__ = [
    "get",
    "CRIDError",
    "parse",
]



from datetime import datetime, timezone
from pathlib import PurePath

from uritools import urisplit  # type: ignore

from ..util import get_session

__LIBRETIME_INFOV2_URL = (
    "https://airtime.service.int.rabe.ch/api/live-info-v2/format/json"
)


def get_show(future: datetime) -> str:  # pragma: no cover
    """Return the slug for a show from LibreTime if it is in the next 7 days.

    Only returns a show for the next seven days because everything futher than
    that is considered unreliable as of early 2023.

    Parameters:
        future: Date to get the show name for.

    Returns:
        Name of the show scheduled for `future`.
    """

    _resp = get_session().get(
        __LIBRETIME_INFOV2_URL,
        params={
            "days": 7,
            "shows": 7000,
        },
        timeout=10,
    )

    _json = _resp.json()
    _tz = _json.get("station").get("timezone")
    _next = _json.get("shows").get("next")
    for _show in _next:
        _start = datetime.fromisoformat(_show.get("starts")).replace(
            tzinfo=timezone.utc
        )
        _end = datetime.fromisoformat(_show.get("ends")).replace(tzinfo=timezone.utc)
        if _start <= future <= _end:
            _path = PurePath(urisplit(_show.get("url")).path)
            return _path.stem

    return ""



from datetime import datetime, timezone
from typing import Optional

from .lib import CRID, canonicalize_show
from .strategy import future, now, past


def get(timestamp: Optional[datetime] = None, fragment: str = "") -> CRID:
    """Get a RaBe CRID.

    Examples:
        You can get a CRID for a specific time.

        ```python
        >>> from datetime import datetime
        >>> from pytz import timezone
        >>> crid = get(datetime(2020, 3, 1, 00, 00, tzinfo=timezone('Europe/Zurich')))
        >>> print(f"version: {crid.version}, start: {crid.start}")
        version: v1, start: ...

        ```

    Parameters:
        timestamp: Exact time you want a CRID for.
            If left empty, a CRID for the current time is generated.
        fragment: Optional fragment to add to the end of the CRID.

    Returns:
        CRID: The generated CRID.
    """
    _now = datetime.now(timezone.utc)
    _ts = timestamp or _now
    _show = ""
    if _ts == _now:
        _show = now.get_show()
    elif _ts < _now:
        _show = past.get_show(past=_ts)
    elif _ts > _now:  # pragma: no cover
        _show = future.get_show(future=_ts)

    if _show:
        _show = canonicalize_show(_show)

    _tscode = f"t=clock={_ts.strftime('%Y%m%dT%H%M%S.%f')[:-4]}Z"
    _fragment = f"{_tscode}{'&' + fragment if fragment else ''}"

    return CRID(f"crid://rabe.ch/v1{'/' + _show if _show else ''}#{_fragment}")



from requests import Session
from requests.adapters import HTTPAdapter, Retry


def get_session(
    retries: int = 5,
    backoff_factor: float = 0.1,
) -> Session:
    """Get a requests session with retry/backoff.

    Parameters:
        retries: How often to retry.
        backoff_factor: A backoff factor to apply between attempts after the
            second try (most errors are resolved immediately by a second try
            without a delay). urllib3 will sleep for::
                {backoff factor} * (2 ** ({number of total retries} - 1))
            seconds. If the backoff_factor is 0.1, then :func:`Retry.sleep`
            will sleep for [0.0s, 0.2s, 0.4s, ...] between retries. It will
            never be longer than `backoff_max`.
            By default, backoff is set to 0.1.
    """
    session = Session()
    session.mount(
        "https://",
        HTTPAdapter(
            max_retries=Retry(
                total=retries,
                backoff_factor=backoff_factor,
            )
        ),
    )
    return session



from datetime import datetime
from pathlib import PurePath
from typing import Optional
from urllib.parse import parse_qs

from slugify import slugify
from uritools import uricompose, urisplit  # type: ignore


def canonicalize_show(show: str) -> str:
    """Get the slug for a show.

    Uses [python-slugify](https://github.com/un33k/python-slugify).

    Parameters:
        show: Name of show with non-ascii chars.
    Returns:
        slugified show name.
    """
    return slugify(show)


class CRIDError(Exception):
    """Represent all cridlib errors."""


class CRIDSchemeMismatchError(CRIDError):
    """Scheme in URI does not match 'crid'."""


class CRIDSchemeAuthorityMismatchError(CRIDError):
    """Hostname part of URI does not match 'rabe.ch'."""


class CRIDUnsupportedVersionError(CRIDError):
    """Unsupported version in path of URI."""


class CRIDMissingMediaFragmentError(CRIDError):
    """Missing media-fragment with clock code."""


class CRIDMalformedMediaFragmentError(CRIDError):
    """Missing media-fragment with clock code."""


CRIDPath = PurePath


class CRID:
    """Represent CRIDs and can parse, validate and render them.

    Examples:
        Generate a CRID from an URL and render it's repr:
        ```python
        >>> CRID("crid://rabe.ch/v1/test#t=clock=19930301T131200.00Z")
        <class 'cridlib.lib.CRID' for 'crid://rabe.ch/v1/test#t=clock=19930301T131200.00Z'>

        ```

        Generate a CRID and render it as str:
        ```python
        >>> str(CRID("crid://rabe.ch/v1/test#t=clock=19930301T131200.00Z"))
        'crid://rabe.ch/v1/test#t=clock=19930301T131200.00Z'

        ```
    """  # noqa: E501

    def __init__(self, uri: Optional[str] = None) -> None:
        """
        Parameters:
            uri: CRID URL to base the new CRID off of.
        """
        self._show: Optional[str] = None
        self._start: Optional[datetime] = None

        self._uri = urisplit(uri)
        if self.scheme != "crid":
            raise CRIDSchemeMismatchError(self.scheme, uri)
        if self.authority != "rabe.ch":
            raise CRIDSchemeAuthorityMismatchError(self.authority, uri)
        # parent.stem contains version in /v1/foo paths, stem in generic root /v1 path
        if self.path.parent.stem != "v1" and self.path.stem != "v1":
            raise CRIDUnsupportedVersionError(self.path, uri)
        self._version = self.path.parent.stem or self.path.stem
        # only store show if we have one
        if self.path.stem != "v1":
            self._show = self.path.relative_to(self.path.parent).stem
        # fragments are optional, but if provided we want them to contain t=code
        if self.fragment:
            try:
                self._start = datetime.strptime(
                    parse_qs(parse_qs(self.fragment)["t"][0])["clock"][0],
                    "%Y%m%dT%H%M%S.%fZ",
                )
            except KeyError as ex:
                raise CRIDMissingMediaFragmentError(self.fragment, uri) from ex
            except ValueError as ex:  # pragma: no cover
                raise CRIDMalformedMediaFragmentError(self.fragment, uri) from ex

    def __str__(self) -> str:
        """
        Returns:
            CRID URL  rendered as string.
        """
        return uricompose(*self._uri)

    def __repr__(self) -> str:
        _fqcn = f"{self.__class__.__module__}.{self.__class__.__qualname__}"
        return f"<class '{_fqcn}' for '{str(self)}'>"

    @property
    def scheme(self) -> str:
        """
        Returns:
            Scheme part of the CRID.
        """
        return self._uri.scheme

    @property
    def authority(self) -> str:
        """
        Returns:
            Authority part (aka hostname) of CRID.
        """
        return self._uri.authority

    @property
    def path(self) -> CRIDPath:
        """
        Returns:
            Path part of CRID.
        """
        return CRIDPath(self._uri.path)

    @property
    def fragment(self) -> str:
        """
        Returns:
            Framgment part of CRID.
        """
        return self._uri.fragment

    @property
    def version(self) -> str:
        """
        Returns:
            Version from CRIDs path.
        """
        return self._version

    @property
    def show(self) -> Optional[str]:
        """
        Returns:
            Show slug from CRIDs path.
        """
        return self._show

    @property
    def start(self) -> Optional[datetime]:
        """
        Returns:
            Start time form CRIDs media fragment.
        """
        return self._start



from .lib import CRID


def parse(value: str) -> CRID:
    """Get CRID dataclass from CRID.

    Parameters:
        value: CRID URL as a string.
    Returns:
        CRID: The parsed CRID.
    """
    return CRID(value)



import xml.etree.ElementTree as ET
from pathlib import PurePath

from uritools import urisplit  # type: ignore

from ..util import get_session

__SONGTICKER_URL = "https://songticker.rabe.ch/songticker/0.9.3/current.xml"


def get_show() -> str:
    """Return the currently running show.

    Calls the the [nowplaying](https://github.com/radiorabe/nowplaying)
    songticker's API.

    Returns:
        Name of the currently running show.
    """

    _resp = get_session().get(__SONGTICKER_URL, timeout=10)
    _tree = ET.fromstring(_resp.text)
    _path = PurePath(urisplit(_tree[3][1].text).path)
    return _path.stem



from datetime import datetime

from ..util import get_session

__ARCHIV_BROADCASTS_URL = "https://archiv.rabe.ch/api/broadcasts/"


def get_show(past: datetime) -> str:
    """Return a show from the past.

    Asks the the [raar](https://github.com/radiorabe/raar) archive for the info.

    Parameters:
        past: Date to get the show name for.

    Returns:
        Show name from the archive for `past`.
    """

    _url = f"{__ARCHIV_BROADCASTS_URL}{past.year}/{past.month:02d}/{past.day:02d}/{past.hour:02d}{past.minute:02d}{past.second:02d}"  # noqa: E501
    _resp = get_session().get(_url, timeout=10)
    _json = _resp.json()
    _data = _json.get("data")
    _label = str(_data[0].get("attributes").get("label"))

    return _label.lower().replace(" ", "-")
